/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "date.h"
#include "util.h"

#define THE_YEAR_0		1970L
#define IS_LEAP_YEAR(y)		(((y) % 4) == 0)


/* returns the seconds east of UTC given `g' and its
   corresponding gmtime() representation.
   This is lifted from mutt 1.2.4, which is GPL and
   copyrighted by Michael R. Elkins <me@cs.hmc.edu> */
static time_t
compute_tz (time_t g, struct tm *utc)
{
	struct tm *lt = localtime (&g);
	time_t t;
	int yday;
	t = (((lt->tm_hour - utc->tm_hour)*60) + (lt->tm_min - utc->tm_min))*60;

	if ((yday = (lt->tm_yday - utc->tm_yday)))
	{
		/* This code is optimized to negative timezones (West of Greenwich) */
		if (yday == -1 ||   /* UTC passed midnight before localtime */
			yday > 1)       /* UTC passed new year before localtime */
			t -= 24 * 60 * 60;
		else
			t += 24 * 60 * 60;
	}

	return t;
}

/* converts struct tm to time_t, but does not take the local
   timezone into account unless ``local'' is nonzero.
   This is lifted from mutt 1.2.4, which is GPL and
   copyrighted by Michael R. Elkins <me@cs.hmc.edu> */
static time_t
mutt_mktime (struct tm *t, gboolean local)
{
	time_t g;

	static int AccumDaysPerMonth[12] = {
		0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334
	};

	/* Compute the number of days since January 1 in the same year */
	g = AccumDaysPerMonth [t->tm_mon % 12];

	/* The leap years are 1972 and every 4. year until 2096,
	 * but this algoritm will fail after year 2099 */
	g += t->tm_mday;
	if ((t->tm_year % 4) || t->tm_mon < 2)
		g--;
	t->tm_yday = g;

	/* Compute the number of days since January 1, 1970 */
	g += (t->tm_year - 70) * 365;
	g += (t->tm_year - 69) / 4;

	/* Compute the number of hours */
	g *= 24;
	g += t->tm_hour;

	/* Compute the number of minutes */
	g *= 60;
	g += t->tm_min;

	/* Compute the number of seconds */
	g *= 60;
	g += t->tm_sec;

	if (local)
		g -= compute_tz (g, t);

	return g;
}

/**
***
**/

typedef struct
{
	char name[5];
	int tz;
}
Timezone_Table_Type;


static Timezone_Table_Type Timezone_Table[] = {
	{"EDT", -500},
	{"PST", -800},
	{"CST", -600},
	{"MST", -700},
	{"MET", 100},				/* Middle European */
	{"MEZ", 100},				/* Middle European */
	{"MSK", 300},				/* Moscow */
	{"CET", 100},				/* Central European */
	{"HKT", 800},				/* Central European */
	{"JST", 900},				/* Central European */
	{"CAST", 930},				/* Central Autsralian */
	{"EAST", 1000},				/* Eastern Autsralian */
	{"NZST", 1200},				/* New Zealand Autsralian */
	{"EET", 200},				/* Eastern European */
	{"", 0}
};


static const gchar*
skip_whitespace (const gchar *b)
{
	while (b!=NULL && (*b==' ' || *b=='\t' || *b=='\n'))
		++b;

	return b;
}


static int
parse_timezone (const gchar *t)
{
	gchar ch;
	int retval = 0;
	Timezone_Table_Type * table = Timezone_Table;

	while (0 != (ch = table->name[0])) {
		if ((*t == ch) && (0 == strncmp(t, table->name, sizeof(table->name)))) {
			retval = table->tz;
			break;
		}
		++table;
	}

	return retval;
}

time_t
rfc822_date_parse (const char *date)
{
	int sign;
	long day;
	long month;
	long year;
	long minutes;
	long seconds;
	long retval;
	long hours;
	long tz;
	char * pch;

	date = skip_whitespace(date);
	if ( !date || !*date )
		return 0;

	/* Look for a weekday, if found skip it */
	while (isalpha((int)*date))
		++date;
	if (*date == ',')
		++date;

	date = skip_whitespace(date);

	/* expecting "03 Feb 1997 10:09:58 +0100" */
	for (day=0; isdigit((int)*date); ++date)
		day = day*10 + (long)(*date - '0');
	if (day<0 || day>31) return 0; --day; /* zero-based */

	/* get month */
	date = skip_whitespace(date); month = 0;
	if (!g_strncasecmp(date,"jan",3)) month = 0;
	else if (!g_strncasecmp(date,"feb",3)) month = 1;
	else if (!g_strncasecmp(date,"mar",3)) month = 2;
	else if (!g_strncasecmp(date,"apr",3)) month = 3;
	else if (!g_strncasecmp(date,"may",3)) month = 4;
	else if (!g_strncasecmp(date,"jun",3)) month = 5;
	else if (!g_strncasecmp(date,"jul",3)) month = 6;
	else if (!g_strncasecmp(date,"aug",3)) month = 7;
	else if (!g_strncasecmp(date,"sep",3)) month = 8;
	else if (!g_strncasecmp(date,"oct",3)) month = 9;
	else if (!g_strncasecmp(date,"nov",3)) month = 10;
	else if (!g_strncasecmp(date,"dec",3)) month = 11;
	else return 0;

	/* skip past month onto year. */
	while (isalpha((int)*date)) ++date;
	date = skip_whitespace(date);

	/* parse the year. */
	pch = NULL;
	year = strtol (date, &pch, 10);
	if (year==0 && pch!=date) /* "25 Apr 00" reported by Jeff Shepherd */
		year = 2000;
	else if (year<100) /* "25 Apr 99" */
		year += 1900;
	else if (year<1900) /* what th..? */
		pan_warn_if_reached ();
	date = skip_whitespace (pch);


	/* Now parse hh:mm:ss */
	for (hours=0; isdigit((int)*date); ++date)
		hours = 10*hours + (long)(*date - '0');
	if (*date!= ':' || hours<0 || hours>=24)
		return 0;
	++date;

	for (minutes=0; isdigit((int)*date); ++date)
		minutes = 10*minutes + (long)(*date-'0');
	if (minutes>=60)
		return 0;

	/* (seconds might not be present) */
	seconds = 0;
	if (*date == ':') {
		for (++date; isdigit((int)*date); ++date)
			seconds = 10*seconds + (long)(*date-'0');
		if (seconds>=60)
			return 0;
	}

	/* Now timezone */
	date = skip_whitespace(date);

	seconds = 0;
	sign = 1;
	if (*date == '+')
		++date;
	else if (*date == '-') {
		sign = -1;
		++date;
	}
	for (tz=0; isdigit((int)*date); ++date)
		tz = tz*10 + (long)(*date-'0');
	tz *= sign;

	date = skip_whitespace(date);
	if (isalpha((int)*date)) {
		sign = 1;
		if ((*date != 'G') && (*date != 'M') && (*date != 'T'))
			tz = parse_timezone(date);
	}
	/* Compute the number of days since beginning of year. */
	day = 31 * month + day;
	if (month > 1) {
		day -= (month * 4 + 27) / 10;
		if (IS_LEAP_YEAR(year))
			++day;
	}
	/* add that to number of days since beginning of time */
	year -= THE_YEAR_0;
	day += year*365 + year/4;
	if ((year%4)==1)
		++day;

	/* Adjust hours for timezone */
	hours -= tz / 100;
	minutes -= tz % 100; /* ?? */

	/* Now convert to secs */
	retval = seconds + 60l*(minutes + 60l*(hours + 24l*day));

	return retval;
}


/*---[ generate_date ]------------------------------------------------
 * generate a properly formatted Date: for a message being sent to
 * an NNTP server.  This follows the format specified by _Son of 1036_
 * section 5.1.  --csk
 * http://www.chemie.fu-berlin.de/outerspace/netnews/son-of-1036.html#5.1
 *--------------------------------------------------------------------*/
gchar*
rfc822_date_generate (time_t secs)
{
	static const gchar *const months[] = {
		"Jan","Feb","Mar","Apr","May","Jun",
		"Jul","Aug","Sep","Oct","Nov","Dec"
	};
	static const gchar *const days[] = {
		"Sun","Mon","Tue","Wed","Thu","Fri","Sat"
	};

	int diff;
	gchar fmt[64] = { '\0' };
	gchar buf[64] = { '\0' };
	struct tm ut;
	struct tm local;

	/* if 0 passed in, use current time */
	if (secs == 0)
		secs = time (NULL);

	/* figure out the difference between this timezone and UT */
	localtime_r (&secs, &local);
	gmtime_r (&secs, &ut);
	diff = (int) difftime (mutt_mktime(&local,0), mutt_mktime(&ut,0));

	/* decide on the time string format.  see son of 1036 sec 5.1 for
	 * the rationale of the offset-from-ut notation.
         *
	 * Update: INN 1.7.2-4 (and others) don't accept the <TZ> string at
	 * the end of son-of-1036's date string.  Thanks to Valentini Jove
	 * for testing to see what worked & what didn't.
	 *
	 * Update 2: Big Thanks to tov are jacobsen <tovj@stud.ntnu.no> for
	 * investigating what format other programs use, for suggesting
	 * code, and for noting that we need to roll our own day-of-week and
	 * month abbreviations because son-of-1036 requires English text but
	 * strftime() localizes (Okt instead of Oct or Man instead of Mon).
	 */
	sprintf (fmt, "%s, %%d %s %%Y %%H:%%M:%%S %c%02d%02d",
		days[local.tm_wday],
		months[local.tm_mon],
		(diff<0?'-':'+'),
		(abs(diff)/3600),
		(abs(diff)%3600));

	/* generate the time */
	strftime (buf, sizeof(buf), fmt, &local);
	return g_strdup (buf);
}
