/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <config.h>
#include <errno.h>

#include <gnome.h>

#include "articlelist.h"
#include "dialogs/dialogs.h"
#include "file-grouplist.h"
#include "fnmatch.h"
#include "globals.h"
#include "grouplist.h"
#include "gui.h"
#include "prefs.h"
#include "queue.h"
#include "task-headers.h"
#include "task-grouplist.h"
#include "util.h"

#include "xpm/folder.xpm"
#include "xpm/envelope.xpm"


static void grouplist_update_ui_thread (void* data);

static void group_dialog_clicked_cb (GnomeDialog*, int, Server*);
static void grouplist_add_group (const Group*);
static void grouplist_update_selected_rows (void);
static void grouplist_grep (GtkWidget *group_entry);

extern GtkWidget   * ttips;

static GnomePixmap * folder_pixmap        = NULL;
static GnomePixmap * group_pixmap         = NULL;

static int           group_mode           = GROUP_ALL;
static GtkWidget   * group_clist_menu     = NULL;
static GtkWidget   * folder_clist_menu    = NULL;
static Server      * my_server            = NULL;
static GtkWidget   * grouplist_mode_menu  = NULL;

PanCallback * grouplist_group_selection_changed = NULL;


/**
 * Returns true if the group deserves to be on the screen based on the
 * grouplist's current server and filter settings.
 * @param group
 * @return true if the group should be in the grouplist
 */
static gboolean
should_be_in_grouplist (const Group * group)
{
	g_return_val_if_fail (group!=NULL, FALSE);
	g_return_val_if_fail (group->server!=NULL, FALSE);

	if (group->server != my_server)
		return FALSE;

	if (group_is_folder(group) && group_mode!=GROUP_FOLDERS)
		return FALSE;

	if (group_mode!=GROUP_ALL && !(group_mode & group->flags))
		return FALSE;

	return TRUE;
}


Server *
grouplist_get_server (void)
{
	return my_server;
}

int
grouplist_get_view_mode (void)
{
	return group_mode;
}

static void
group_dialog_clicked_cb (GnomeDialog  * dialog,
                         int            num,
                         Server       * server)
{
	switch (num)
	{
		case 0: /* OK button */
			queue_add (TASK(task_grouplist_new (
						server, TRUE, GROUPLIST_ALL)));
		default:
			gtk_widget_destroy (GTK_WIDGET(dialog));
			break;
	}
}                                                                               

void
grouplist_set_server (Server * server)
{
	g_return_if_fail (server != NULL);

	/* nothing to do */
	if (my_server == server)
		return;

	/* Clear the ui and set the active list */
	gtk_clist_clear (GTK_CLIST(Pan.group_clist));
	my_server = server;

	/* already loaded the groups */
	if (server->groups->len) {
		grouplist_update_ui (server);
		return;
	}

	/* haven't loaded the groups, but found a file for it */
	if (file_grouplist_exists (server)) {
		grouplist_load (server);
		return;
	}

	/* can't even find a file for the groups;
	   shall we download a group list? */
	{
		GtkWidget * d, * l;
		gchar * s;
		d = gnome_dialog_new (_("No Group List"),
	                              GNOME_STOCK_BUTTON_YES,
	                              GNOME_STOCK_BUTTON_NO,
	                              NULL);
		s = g_strdup_printf (_("We don't have a group list for server "
		                       "\"%s\".\nShall we get one?\n"),
		                       server->name);
		l = gtk_label_new (s);
		g_free (s);
		gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(d)->vbox), l);
		gtk_signal_connect (GTK_OBJECT(d), "clicked",
		                    GTK_SIGNAL_FUNC (group_dialog_clicked_cb),
				    server);
		gui_popup_draw (GTK_WIDGET(d), Pan.window);
	}
}

/* process what happens when a user lcicks on a group in the group list */
static void
select_row_cb (GtkWidget  * widget,
               int          row,
               int          col,
               GdkEvent   * event)
{
	GdkEventButton * bevent = (GdkEventButton *)event;
	if (bevent == NULL)
	       return;

	pan_callback_call (grouplist_group_selection_changed,
	                   Pan.group_clist,
	                   NULL);

	if (bevent->button==1 && bevent->type==GDK_2BUTTON_PRESS)
	{
		/* which group? */
		gpointer * foo = gtk_clist_get_row_data(GTK_CLIST(widget), row);
		Group* group = GROUP(foo);

		/* turn the page if in notebook layout. */
		gui_page_set (ARTICLELIST_PAGE, Pan.article_ctree);

		/* load the specified group */
		if (articlelist_get_group() != group)
			articlelist_set_group (group);
	}
}

static void
grouplist_process_selected (int type)
{
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	const gboolean has_selection = clist->selection != NULL;
	Group * group = NULL;
	GList * list = NULL;
	GList * l = NULL;

	/* build the list of groups */
	if (has_selection) {
		list = g_list_copy (GTK_CLIST (Pan.group_clist)->selection);
		for (l=list; l!=NULL; l=l->next)
			l->data = gtk_clist_get_row_data (
				GTK_CLIST (Pan.group_clist),
				GPOINTER_TO_INT (l->data));
	}

	if (!list && (group=articlelist_get_group()))
	{
		switch (type)
		{
			case 3:
				queue_add (TASK(task_headers_new(
					group, HEADERS_ALL, TRUE)));
				break;
			case 4:
				queue_add (TASK(task_headers_new(
					group, HEADERS_NEW, TRUE)));
				break;
			default:
				break;
		}
	}

	for (l=list; l!=NULL; l=l->next)
	{
		group = GROUP(l->data);

		switch (type)
		{
			case 1: /* subscribe */
				group_set_subscribed (group, TRUE);
				break;

			case 2: /* unsubscribe */
				group_set_subscribed (group, FALSE);
				break;
				break;

			case 3: /* download all */
				queue_add (TASK(task_headers_new(
					group, HEADERS_ALL, FALSE)));
				break;

			case 4: /* download new */
				queue_add (TASK(task_headers_new(
					group, HEADERS_NEW, FALSE)));
				break;

			case 5: /* clean/empty */
				group_empty_dialog (group);
				break;

			case 6: /* group properties */
				dialog_group_props (group);
				break;

			case 10: /* download all bodies */
				group_download_all_articles (group);
				break;

			default:
				break;
		}
	}

	/* if unsubscribe and in subscribe mode, remove the groups */
	if (type==2 && group_mode==GROUP_SUBSCRIBED)
		for (l=list; l!=NULL; l=l->next)
			grouplist_remove_row (GROUP(l->data));

	/* refresh the modified entries */
	grouplist_update_selected_rows ();

	/* unselect */
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));

	/* cleanup */
	g_list_free (list);
}

/**
***
**/

GPtrArray*
grouplist_get_subscribed_groups (void)
{
	GPtrArray * a = g_ptr_array_new ();

	if (my_server != NULL)
	{
		guint i;
		pan_g_ptr_array_reserve (a, my_server->groups->len);
		for (i=0; i!=my_server->groups->len; ++i)
		{
			Group * g = GROUP(g_ptr_array_index(my_server->groups,i));
			if (group_is_subscribed (g))
				g_ptr_array_add (a, g);
		}
	}

	return a;
}

Group*
grouplist_get_selected_group (void)
{
	GList * l;
	GtkCList * clist;
	Group * retval = NULL;
       
	clist = GTK_CLIST(Pan.group_clist);
	l = clist->selection;
	if (l != NULL)
		retval = GROUP(gtk_clist_get_row_data(clist, GPOINTER_TO_INT(l->data)));

	return retval;
}

GPtrArray*
grouplist_get_selected_groups (void)
{
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GPtrArray * retval = NULL;
	GList * l = NULL;

	/* initialize the gptrarray */
       	retval = g_ptr_array_new ();
	pan_g_ptr_array_reserve (retval, g_list_length(clist->selection));

	/* get the group selected items; otherwise, 
	   use the articlelist's group */
	for (l=clist->selection; l!=NULL; l=l->next)
		g_ptr_array_add (retval, gtk_clist_get_row_data (clist, GPOINTER_TO_INT (l->data)));
	if (!retval->len) {
		Group * group = articlelist_get_group ();
		if (group != NULL)
			g_ptr_array_add (retval, group);
	}

	return retval;
}

/**
***
**/

void
grouplist_selected_subscribe (void)
{
	grouplist_process_selected (1);
}
void
grouplist_selected_unsubscribe (void)
{
	grouplist_process_selected (2);
}
void
grouplist_selected_empty (void)
{
	grouplist_process_selected (5);
}
void
grouplist_selected_properties (void)
{
	grouplist_process_selected (6);
}

void
grouplist_selected_update_count_info (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	if (a != NULL)
		group_update_count_info (a);
}

void
grouplist_selected_catchup (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	if (a != NULL)
		pan_g_ptr_array_foreach (a, (GFunc)group_catchup, NULL);
	g_ptr_array_free (a, TRUE);
}



static void
grouplist_selected_rename_cb (gchar* string, gpointer data)
{
	if (is_nonempty_string(string))
	{
		Group * group = GROUP(data);
		group_set_string_field (group, &group->readable_name, string);
		group_set_dirty (group);
		grouplist_update_row (group);
	}
}
void
grouplist_selected_rename (void)
{
	Group * group = grouplist_get_selected_group ();
	if (group != NULL)
	{
		gchar * fmt = group_is_folder(group)
			? _("Rename Folder \"%s\"")
			: _("Rename Group \"%s\"");
		gchar * prompt = g_strdup_printf (fmt,
			group_get_readable_name(group));
		gnome_request_dialog (FALSE,
		                      prompt,
		                      group->name,
			              128,
			              grouplist_selected_rename_cb,
			              group,
			              GTK_WINDOW(Pan.window));
		g_free (prompt);
	}
}

void
grouplist_selected_destroy (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	if (a->len != 0)
	{
		guint i;
		guint button;
		GString * str = g_string_new (NULL);

		/* build the warning string */
		g_string_assign (str, _("Make sure you really want to do this:\n"
		                        "This operation will delete the folder and its contents!\n"
				       	" \n"
					"Do you REALLY want to delete:\n \n"));
		for (i=0; i!=a->len; ++i)
			g_string_sprintfa (str, "     %s\n", GROUP(g_ptr_array_index(a,i))->name);

		/* ask the user if they *REALLY* want to do this */
		button = gnome_dialog_run_and_close (GNOME_DIALOG(
				gnome_message_box_new (str->str,
					GNOME_MESSAGE_BOX_QUESTION,
					GNOME_STOCK_BUTTON_YES,
					GNOME_STOCK_BUTTON_NO,
					NULL)));

		/* if they really wnat to do it, destroy the groups */
		if (button==0) {
			for (i=0; i!=a->len; ++i) {
				Group * g = GROUP(g_ptr_array_index(a,i));
				server_destroy_group (g->server, g);
			}
		}

		g_string_free (str, TRUE);
	}
	g_ptr_array_free (a, TRUE);
}

/**
***  Process all the commands to download
***  [new | all | sample] [articles | headers]
**/

static void
grouplist_start_task_headers (GPtrArray * a,
                              HeaderDownloadType type,
                              gboolean bodies)
{
	int i;

	/* sample all the selected groups */
	for (i=0; i<a->len; ++i)
	{
		Group * group = GROUP(g_ptr_array_index(a,i));
		PanObject * o = bodies
			? task_headers_new_with_bodies (group, type)
			: task_headers_new (group, type, FALSE);
		queue_add (TASK(o));
	}

	/* cleanup */
	grouplist_update_selected_rows ();
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));
}

void
grouplist_selected_download_dialog (void)
{
	int i;
	GPtrArray * a = grouplist_get_selected_groups ();

	/* sample all the selected groups */
	for (i=0; i<a->len; ++i) {
		Group * group = GROUP(g_ptr_array_index(a,i));
		if (!group_is_folder(group))
			dialog_download_headers (group);
	}

	/* cleanup */
	grouplist_update_selected_rows ();
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));
	g_ptr_array_free (a, TRUE);
}

void
grouplist_selected_download_sample (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_SAMPLE, FALSE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_selected_download_sample_bodies (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_SAMPLE, TRUE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_selected_download_all (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_ALL, FALSE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_selected_download_all_bodies (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_ALL, TRUE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_selected_download_new (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_NEW, FALSE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_selected_download_new_bodies (void)
{
	GPtrArray * a = grouplist_get_selected_groups ();
	grouplist_start_task_headers (a, HEADERS_NEW, TRUE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_subscribed_download_new (void)
{
	GPtrArray * a = grouplist_get_subscribed_groups ();
	grouplist_start_task_headers (a, HEADERS_NEW, FALSE);
	g_ptr_array_free (a, TRUE);
}
void
grouplist_subscribed_download_new_bodies (void)
{
	GPtrArray * a = grouplist_get_subscribed_groups ();
	grouplist_start_task_headers (a, HEADERS_NEW, TRUE);
	g_ptr_array_free (a, TRUE);
}

void
grouplist_get_all (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (
			TASK(task_grouplist_new (
					my_server, TRUE, GROUPLIST_ALL)));
}

void
grouplist_get_new (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (
			TASK(task_grouplist_new (
					my_server, TRUE, GROUPLIST_NEW)));
}


void
grouplist_select_all (void)
{
	pan_lock();
	gtk_clist_select_all (GTK_CLIST (Pan.group_clist));
	pan_unlock();
}


static void
grouplist_button_press (GtkWidget *widget, GdkEventButton *bevent)
{
	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
		                              "button_press_event");

	if (bevent->button == 3)
	{
                gboolean selected = FALSE;
		GList * l = NULL;

		/* find what we need to know for setting items insensitive */
		GPtrArray * a = grouplist_get_selected_groups ();
		selected = a->len != 0;
		g_ptr_array_free (a, TRUE);

		if (group_mode == GROUP_FOLDERS) /* We are dealing with folders */
		{
			pan_lock ();
			l=GTK_MENU_SHELL(folder_clist_menu)->children;
			/* l is new folder */
			l=l->next; /* rename group/folder */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* delete folder */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);

			/* show the menu */
			gtk_menu_popup (GTK_MENU(folder_clist_menu),
		                NULL, NULL, NULL, NULL,
		                bevent->button, bevent->time);
			pan_unlock ();
		}
		else /* We must be dealing with real groups */
		{
			pan_lock ();
			l=GTK_MENU_SHELL(group_clist_menu)->children; 
			/* l is get new headers in subscribed */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), TRUE);
			l=l->next; /* l is get new headers in selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* l is get all headers in selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* l is sample for selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* separator */
			l=l->next; /* l is get new headers & bodies in subscribed */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), TRUE);
			l=l->next; /* l is get new headers & bodies in selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* l is get all headers & bodies in selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* l is sample new headers & bodies in selected */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* separator */
			l=l->next; /* refresh article count */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* catch up */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
	
			l=l->next; /* separator */
			l=l->next; /* subscribe */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* unsubscribe */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
	
			l=l->next; /* separator */
			l=l->next; /* empty group */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* rename group */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
			l=l->next; /* properties */
			gtk_widget_set_sensitive (GTK_WIDGET(l->data), selected);
	
			/* show the menu */
			gtk_menu_popup (GTK_MENU(group_clist_menu),
			                NULL, NULL, NULL, NULL,
			                bevent->button, bevent->time);
			pan_unlock ();
		} 
	}
}

void
grouplist_set_view_mode (GtkWidget *widget, gpointer data)
{
	GtkWidget *menu = NULL;

	group_mode = GPOINTER_TO_INT (data);
	grouplist_update_ui (grouplist_get_server());

        pan_lock ();
        menu = gtk_option_menu_get_menu (GTK_OPTION_MENU(grouplist_mode_menu));
        gtk_object_ref (GTK_OBJECT(menu));
        gtk_option_menu_remove_menu (GTK_OPTION_MENU (grouplist_mode_menu));
	switch (group_mode) {
		case GROUP_ALL:
			gtk_menu_set_active (GTK_MENU (menu), 0);
			break;
		case GROUP_SUBSCRIBED:
			gtk_menu_set_active (GTK_MENU (menu), 1);
			break;
		case GROUP_NEW:
			gtk_menu_set_active (GTK_MENU (menu), 2);
			break;
		case GROUP_FOLDERS:
			gtk_menu_set_active (GTK_MENU (menu), 3);
			break;
		default: break;
	}
        gtk_option_menu_set_menu (GTK_OPTION_MENU(grouplist_mode_menu), menu);
        gtk_object_unref (GTK_OBJECT(menu));
        pan_unlock ();
}

typedef struct
{
	gchar * name;
	int mode;
}
GrouplistModeMenuStruct;

static GtkWidget*
grouplist_mode_menu_create (void)
{
	GtkWidget *option_menu = gtk_option_menu_new();
	GtkWidget *menu = gtk_menu_new ();
	int index = 0;
	int i;
	GrouplistModeMenuStruct foo[] = {
		{_("All Groups"), GROUP_ALL},
		{_("Subscribed"), GROUP_SUBSCRIBED},
		{_("New Groups"), GROUP_NEW},
		{_("Folders"), GROUP_FOLDERS}
	};
	const int row_qty = sizeof(foo) / sizeof(foo[0]);

	for (i=0; i<row_qty; ++i) {
		GtkWidget *item = gtk_menu_item_new_with_label (foo[i].name);
		gtk_signal_connect (GTK_OBJECT(item), "activate",
				    (GtkSignalFunc)grouplist_set_view_mode,
				    GINT_TO_POINTER(foo[i].mode));
		gtk_menu_append (GTK_MENU (menu), item);
		if (group_mode == foo[i].mode)
			index = i;
	}

	gtk_menu_set_active (GTK_MENU(menu), index);
        gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
	gtk_widget_show_all (GTK_WIDGET(option_menu));

	return option_menu;
}

/*---[ grouplist_create ]---------------------------------------------
 *
 *--------------------------------------------------------------------*/
GtkWidget *
grouplist_create (void)
{
	GtkWidget * w;
        GtkWidget * vbox;
        GtkWidget * hbox;
	GtkWidget * toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_TEXT);
	GtkCList * clist;

	char *titles[] = {
		"",
		_("Groups"),
		_("Unread"),
		_("Total"),
		_("Description")
	};

	grouplist_group_selection_changed = pan_callback_new ();

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GNOME_PAD_SMALL); 
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);

	group_mode = gnome_config_get_int ("/Pan/State/group_mode=1");
	if ( group_mode!=GROUP_ALL
	     && group_mode!=GROUP_SUBSCRIBED
	     && group_mode!=GROUP_NEW
	     && group_mode!=GROUP_FOLDERS)
		group_mode = GROUP_ALL;
	
	Pan.group_clist = gtk_clist_new_with_titles (5, titles);
	clist = GTK_CLIST(Pan.group_clist);
	widget_set_font (GTK_WIDGET(Pan.group_clist), grouplist_font);
	gtk_clist_set_column_justification (clist, 2, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification (clist, 3, GTK_JUSTIFY_RIGHT);


	w = gtk_label_new (_("Groups"));
	gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar), w, "", "");
	gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

	grouplist_mode_menu = grouplist_mode_menu_create();
	gtk_widget_set (grouplist_mode_menu, "width", 120, "height", 25, NULL); /* FIXME: If we don't do this the om doesn't show up*/
	gtk_box_pack_start (GTK_BOX(hbox), grouplist_mode_menu, FALSE, FALSE, 0);

	w = gtk_label_new (_("Find:"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);

	w = gtk_entry_new ();
	gtk_signal_connect (GTK_OBJECT (w), "activate",
			    GTK_SIGNAL_FUNC(grouplist_grep), w);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w,
		_("Type in a group search string and press ENTER.  "
		  "Wildcards are allowed."), "");
	gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 100);
	gtk_widget_show_all (toolbar);
	gtk_box_pack_start (GTK_BOX(hbox), w, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	w = gtk_scrolled_window_new (NULL, NULL);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (w),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(w), Pan.group_clist);
        gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "select_row",
			    GTK_SIGNAL_FUNC (select_row_cb), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "unselect_row",
			    GTK_SIGNAL_FUNC (select_row_cb), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "button_press_event",
			    GTK_SIGNAL_FUNC (grouplist_button_press), NULL);

	gtk_clist_set_column_width     (clist, 0, 16);
	gtk_clist_set_column_min_width (clist, 0, 16);
	gtk_clist_set_column_max_width (clist, 0, 16);

	gtk_clist_set_column_width (clist, 1, 170);
	gtk_clist_set_column_width (clist, 2, 36);
	gtk_clist_set_column_width (clist, 3, 33);
	gtk_clist_set_column_width (clist, 4, 350);

/* JEL: this works well in paned mode, but not so well in notebook mode, we
   need to find a solution that will cover both ... maybe splitting the justification
   parts into the _construct funcs for both ...
	gtk_clist_set_column_justification (clist, 0, GTK_JUSTIFY_LEFT );
	gtk_clist_set_column_justification (clist, 1, GTK_JUSTIFY_RIGHT );
	gtk_clist_set_column_justification (clist, 2, GTK_JUSTIFY_RIGHT );
*/
	gtk_clist_set_selection_mode (clist, GTK_SELECTION_EXTENDED);

	/* Create the Popup Menu while we are here. */
	group_clist_menu = gnome_popup_menu_new (group_menu);	
	folder_clist_menu = gnome_popup_menu_new (folder_menu);	

	/* Lets have the Pixmap ready too */
	folder_pixmap = (GnomePixmap*) gnome_pixmap_new_from_xpm_d (folder_xpm);
	group_pixmap = (GnomePixmap*) gnome_pixmap_new_from_xpm_d (envelope_xpm);

	gtk_widget_show_all (vbox);
	return vbox;
}


/* Grep the group listing for whatever the user typed into the Find box */
static void
grouplist_grep (GtkWidget *group_entry)
{
	int i;
	gchar * s;
	gchar * pch;
	Server * server = grouplist_get_server ();

	g_return_if_fail (Pan.group_clist);
	g_return_if_fail (server);

	pan_lock ();

	/* wrap the key in wildcards to make it a substring search...
	   unless there are already wildcards present, indicating the
	   user already knows what he wants */
        pch = gtk_editable_get_chars (GTK_EDITABLE(group_entry), 0, -1);
	pch = g_strstrip (pch);
	if (strchr(pch,'*'))
		s = g_strdup (pch);
	else if (*pch)
		s = g_strdup_printf("*%s*",pch);
	else
		s = NULL;
	g_free (pch);
	pch = NULL;
g_message ("grp string is [%s]", pch?pch:"(Null)");

	gtk_clist_freeze (GTK_CLIST(Pan.group_clist));
	gtk_clist_clear (GTK_CLIST(Pan.group_clist));
	for (i=server->groups->len-1; i>=0; --i)
	{
		Group * group = GROUP(g_ptr_array_index (server->groups, i));
		if (!should_be_in_grouplist (group))
			continue;
		if (is_nonempty_string(s) && fnmatch(s,group->name,0))
			continue;

		grouplist_add_group (group);

	}
	gtk_clist_thaw (GTK_CLIST(Pan.group_clist));
	pan_unlock();

	g_free (s);
}


static gchar*
grouplist_load_describe (const StatusItem* item)
{
	return g_strdup (_("Loading Groups"));
}

/* Load a grouplist database into memory (in a separate thread) */
static void
grouplist_load_thread (Server * server)
{
	StatusItem * status = NULL;

	/* create a new status item to get load messages */
	status = STATUS_ITEM(status_item_new(grouplist_load_describe));
	pan_object_ref(PAN_OBJECT(status));
	pan_object_sink(PAN_OBJECT(status));
	gui_add_status_item (status);

	/* load the groups... */
	file_grouplist_load (server, status);

	/* remove the "load groups" status: grouplist_update_ui has its own */
	gui_remove_status_item (status);
	pan_object_unref(PAN_OBJECT(status));
	status = NULL;

	/* update the grouplist ui */
	grouplist_update_ui_thread (server);
}

/**
 * grouplist_load:
 * @server: The server of the grouplist to load
 *
 * Load a grouplist database into memory
 **/
void
grouplist_load (Server *server)
{
	pthread_t thread;
	pthread_create (&thread, NULL, (void*)grouplist_load_thread, server);
	pthread_detach (thread);
}



/**
***
**/

void
grouplist_set_selected_groups (const GPtrArray * a)
{
	GtkCList * clist;

	if (a != NULL)
	{
		guint i;

		clist = GTK_CLIST(Pan.group_clist);
		pan_lock ();
		gtk_clist_freeze (clist);
		gtk_clist_unselect_all (clist);
		for (i=0; i!=a->len; ++i)
		{
			const Group * g = GROUP(g_ptr_array_index(a,i));
			const int row = gtk_clist_find_row_from_data (clist, (gpointer)g);
			gtk_clist_select_row (clist, row, 0);
			if (!gtk_clist_row_is_visible (clist, row))
			{
				gfloat valign = 0.5;
				gfloat halign = 0.0;
				gtk_clist_moveto (clist, row, 0, (double)valign, (double)halign);
			}
		}
		gtk_clist_thaw (clist);
		pan_unlock ();
	}
}

/**
***
**/

/*---[ grouplist_update ]---------------------------------------------
 * update the entire group CLIST, clearing and then filling it in
 * with the groups that are part of the current group mode (all, 
 * subscribed, etc.)
 *--------------------------------------------------------------------*/

static gchar*
grouplist_update_describe (const StatusItem *item)
{
	return g_strdup (_("Updating Group List"));
}

static void
grouplist_update_ui_thread (void* data)
{
	Server * server = (Server*)data;
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GPtrArray * sel;
	StatusItem *item = NULL;
	int i;

	g_return_if_fail (server != NULL);

        /* create a new status item to get sort/thread messages */
	item = STATUS_ITEM(status_item_new(grouplist_update_describe));
	pan_object_ref(PAN_OBJECT(item));
	pan_object_sink(PAN_OBJECT(item));
	gui_add_status_item (item);
	status_item_emit_init_steps (item, server->groups->len);
	status_item_emit_status (item, _("Updating Group List"));

	pan_lock();

	/* set the groups column title */
	if (group_mode == GROUP_SUBSCRIBED)
		gtk_clist_set_column_title (clist, 1, _("Subscribed Groups"));
	else if (group_mode == GROUP_ALL)
		gtk_clist_set_column_title (clist, 1, _("All Groups"));
	else if (group_mode == GROUP_NEW)
		gtk_clist_set_column_title (clist, 1, _("New Groups"));
	else if (group_mode == GROUP_FOLDERS)
		gtk_clist_set_column_title (clist, 1, _("Folders"));

	/* clear the clist */
	sel = grouplist_get_selected_groups ();
	gtk_clist_freeze (clist);
	gtk_clist_clear (clist);

	/* add the groups/folders */
	if (group_mode == GROUP_FOLDERS)
		server = server_get_by_name(INTERNAL_SERVER_NAME);
	for (i=server->groups->len-1; i>=0; --i) {
		Group* group = GROUP(g_ptr_array_index (server->groups, i));
		if ((group_mode==GROUP_ALL) || (group_mode & group->flags)) {
			grouplist_add_group (group);
			pan_unlock();
			status_item_emit_next_step (item);
			pan_lock();
		}
	}

	gtk_clist_thaw (clist);
	pan_unlock();

	/* update the selection */
	grouplist_set_selected_groups (sel);
	g_ptr_array_free (sel, TRUE);

	/* clean out the status item */
	gui_remove_status_item (item);
	pan_object_unref(PAN_OBJECT(item));
}

void
grouplist_update_ui (Server *server)
{
	pthread_t thread;

	g_return_if_fail (server != NULL);
	pthread_create (&thread, NULL, (void*)grouplist_update_ui_thread, server);
	pthread_detach (thread);
}


/* note that this is called inside a pan_lock() */
static void
grouplist_add_group (const Group * group)
{
	int row;
	GnomePixmap * pixmap;
	char total_buf[16] = { '\0' };
	char unread_buf[16] = { '\0' };
	const gint total = group->article_qty;
	const gint unread = MAX(0, total - group->article_read_qty);
	char * const text[7] = {
		"",
		(gchar*)group_get_readable_name(group),
		unread_buf,
		total_buf,
		(gchar*)(group->description?group->description:""),
		NULL
	};
	commatize_ulong (total, total_buf);
	commatize_ulong (unread, unread_buf);

	/* add the row, in reverse order for speed -- gtkclist
	   is stupid and has to walk through its internal GSList
	   for each insert */
	row = gtk_clist_prepend (GTK_CLIST(Pan.group_clist), (char**)text);

	/* add the pixmap, if any */
	pixmap = NULL;
	if (group_is_folder (group))
		pixmap = folder_pixmap;
	else if (group_is_subscribed (group))
		pixmap = group_pixmap;
	if (pixmap != NULL)
		gtk_clist_set_pixmap (GTK_CLIST(Pan.group_clist), row, 0,
		                      pixmap->pixmap,
		                      pixmap->mask);

	gtk_clist_set_row_data (GTK_CLIST(Pan.group_clist), 0, (gpointer)group);
}

/* Remove a single row from the group CList, useful in unsubscribing */
void
grouplist_remove_row (const Group *group)
{
	if (group->server == my_server)
	{
		GtkCList *list = GTK_CLIST(Pan.group_clist);
		int row;

		pan_lock();
		row = gtk_clist_find_row_from_data (list, (gpointer)group);
		if (row != -1)
			gtk_clist_remove (list, row);
		pan_unlock();
	}
}


/**
 * grouplist_update_row:
 * @group: Group who's row is to be updated.
 *
 * Update the CList row for a group to show it's message counts or
 * any other relative data (subscribed/on disk) for that group.
 **/
void
grouplist_update_row (const Group *group)
{
	char buf[16];
	gint row = 0;
	const gint total = group->article_qty;
	const gint unread = MAX(0, total-group->article_read_qty);
	GtkCList * list = GTK_CLIST (Pan.group_clist);

	pan_lock();

	row = gtk_clist_find_row_from_data (list,(gpointer)group);
	if (row != -1)
	{
		/* icon column */
		if (group_is_subscribed (group))
			gtk_clist_set_pixmap (
				GTK_CLIST(Pan.group_clist), row, 0,
				group_pixmap->pixmap,
				group_pixmap->mask);
		else if (!group_is_folder(group))
			gtk_clist_set_text (GTK_CLIST(Pan.group_clist),
					    row, 0, "");

		/* name column */
		gtk_clist_set_text (list, row, 1,
		                    group_get_readable_name(group));

		/* total column */
		commatize_ulong (total, buf);
		gtk_clist_set_text (list, row, 3, buf);

		/* unread column */
		commatize_ulong (unread, buf);
		gtk_clist_set_text (list, row, 2, buf);
	}
	else if (group->server == my_server) /* not in list */
	{
		if (should_be_in_grouplist (group)) /* possibly a new folder... */
		{
			grouplist_add_group (group); /* FIXME: add sorted */
		}
	}

	pan_unlock();
}

static void
grouplist_update_selected_rows (void)
{
	GtkCList *list = GTK_CLIST(Pan.group_clist);
	GList *l = list->selection;
	for (l=list->selection; l!=NULL; l=l->next)
	{
		Group *group = NULL;
		int index = GPOINTER_TO_INT(l->data);
		group = GROUP(gtk_clist_get_row_data (list, index));
		grouplist_update_row (group);
	}
}

static void
grouplist_internal_select_row (int row)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);

	pan_lock();
	gtk_clist_unselect_all (list);
	gtk_clist_select_row (list,row,-1);
	pan_unlock();
}

#if 0
void
grouplist_prev_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;

	if (!row_qty)
		return;

	if ( list->selection ) {
		row = GPOINTER_TO_INT(list->selection->data) - 1;
		if (row<0)
			row = row_qty-1;
	}

	grouplist_internal_select_row (row);
	articlelist_set_group (GROUP(gtk_clist_get_row_data(list,row)));
}

void
grouplist_next_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;

	if (!row_qty)
		return;

	if ( list->selection )
		row = (GPOINTER_TO_INT(list->selection->data)+1) % row_qty;

	grouplist_internal_select_row (row);
	articlelist_set_group (GROUP(gtk_clist_get_row_data(list,row)));
}
#endif

static gboolean
grouplist_select_row_if_unread (int row)
{
        GtkCList * list = GTK_CLIST(Pan.group_clist);
	Group * g = gtk_clist_get_row_data(list,row);
	const gulong total = g->article_qty;
	const gulong read = g->article_read_qty;
	gboolean retval = FALSE;

	if (read < total)
	{
		grouplist_internal_select_row (row);
		articlelist_set_group (g);
		retval = TRUE;
	}

	return retval;
}

void
grouplist_next_unread_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
	int i;
       	int row = 0;

	if (!row_qty)
		return;

	/* get the first row */
	if ( list->selection )
		row = (GPOINTER_TO_INT(list->selection->data) + 1) % row_qty;
	else
		row = 0;

	/* walk through */
	for ( i=0; i<row_qty; ++i, row=++row%row_qty )
		if ( grouplist_select_row_if_unread (row) )
			return;
}

#if 0
void
grouplist_prev_unread_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;
	int i;

	if (!row_qty)
		return;

	/* get the first row */
	if ( list->selection ) {
		row = GPOINTER_TO_INT(list->selection->data);
		row = row ? row-1 : row_qty-1;
	}
	else
		row = row_qty-1;

	/* walk through */
	for ( i=0; i!=row_qty; ++i, row=row?--row:row_qty-1 )
		if ( grouplist_select_row_if_unread (row) )
			return;
}
#endif
