/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "nntp.h"
#include "task-headers.h"

#include "articlelist.h"
#include "grouplist.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

#define SAMPLE_SIZE_DEFAULT 150

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint task_headers_run (Task* item);

static gchar* task_headers_describe (const StatusItem *item);

static int task_get_header_range (TaskHeaders  * task,
                                  gulong * setme_low,
                                  gulong * setme_high,
                                  gulong * setme_first,
                                  gulong * setme_last,
		                  gulong * setme_qty,
		                  const char ** setme_progress_fmt);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

/*****
******  TASK LIFE CYCLE
*****/

static void
task_headers_destructor (PanObject* o)
{
	TaskHeaders * task_h = TASK_HEADERS(o);

	/* destruct this class: cleanup articles... */
	g_ptr_array_free (task_h->articles, TRUE);
	task_h->articles = NULL;

	/* destruct parent class... */
	task_destructor (o);
}

static PanObject*
real_task_headers_new (Group                * group,
		       HeaderDownloadType     dl_type,
                       gboolean               high_priority,
		       gboolean               download_bodies,
                       int                    sample_size)
{
	TaskHeaders *item;

       	item = g_new0 (TaskHeaders, 1);
	debug1 (DEBUG_PAN_OBJECT, "task_headers_new: %p", item);

	task_constructor (TASK(item),
	                  task_headers_destructor,
	                  task_headers_describe,
	                  task_headers_run,
	                  group->server,
	                  high_priority,
	                  TRUE);
	item->group = group;
	item->download_type = dl_type;
	item->download_bodies = download_bodies;
	item->articles = g_ptr_array_new ();
	item->body_index = 0;
	item->sample_size = sample_size;

	return PAN_OBJECT(item);
}

PanObject*
task_headers_new (Group               * group,
                  HeaderDownloadType    dl_type,
                  gboolean              high_priority)
{
	return real_task_headers_new (group,
	                              dl_type,
	                              high_priority,
	                              FALSE,
	                              SAMPLE_SIZE_DEFAULT);
}

PanObject*
task_headers_new_with_bodies   (Group               * group,
				HeaderDownloadType    dl_type)
{
	return real_task_headers_new (group,
	                              dl_type,
	                              FALSE,
	                              TRUE,
	                              SAMPLE_SIZE_DEFAULT);
}

PanObject*
task_headers_new_sample (Group     * group,
			 guint       sample_size,
                         gboolean    download_bodies)
{
	return real_task_headers_new (group,
	                              HEADERS_SAMPLE,
	                              FALSE,
	                              download_bodies,
	                              sample_size);
}

/*****
******
*****/

static gint
task_headers_run (Task* task)
{
	StatusItem * status = STATUS_ITEM(task);
	TaskHeaders * task_h = TASK_HEADERS(task);
	gboolean active;
	gulong read;
	int i;
	gulong first;
	gulong last;
	gulong low;
	gulong high;
	gulong qty;
	const char * progress_fmt;
	int retval = -1; /* -1 if unset; 0 if success; 1 if fail */
	gboolean reffed = FALSE;

	/* find out how many article headers we're getting */
	low = high = first = last = qty = 0;
	progress_fmt = NULL;
	if (task_get_header_range(task_h,
	                          &low,
	                          &high,
	                          &first,
	                          &last,
	                          &qty,
	                          &progress_fmt))
		retval = 1;

	/* no articles to fetch */
	if (retval==-1 && qty==0)
		retval = 0;

	active = task_h->group==articlelist_get_group ();
	if (retval==-1)
	{
		int steps;

		/* load the existing headers */
		group_ref_articles (task_h->group, status);
		reffed = TRUE;

		/* init the status item for the header download */
		switch (task_h->download_type) {
			case HEADERS_ALL: steps = high-low; break;
			case HEADERS_NEW: steps = high - task_h->group->article_high; break;
			case HEADERS_SAMPLE: steps = task_h->sample_size;
			default: pan_warn_if_reached(); steps = 0;
		}
		status_item_emit_init_steps (status, steps);

		/* get the article headers... */
		i = nntp_articlelist_download (status,
					       task->sock,
					       task_h->group,
					       low,
					       high,
					       &task->hint_abort,
					       TRUE,
					       progress_fmt,
					       task_h->articles);
		if (i)
			retval = i;
	}

	if (retval==-1)
	{
		/* update the article qty for the group */
		group_set_article_range (task_h->group, first, last);
		group_expire_old_articles (task_h->group);

		if (task_h->download_type==HEADERS_ALL)
			group_set_article_qty (task_h->group, task_h->articles->len);
		else
			group_inc_article_qty (task_h->group, high-low);

		/* update the read article qty for the group */
		read = 0;
		for (i=0; i!=task_h->articles->len; ++i)
			if (article_is_read (ARTICLE(
				g_ptr_array_index(task_h->articles,i))))
					++read;
		group_inc_article_read_qty (task_h->group, read);

		/* no articles! */
		if (!task_h->articles->len)
		{
			status_item_emit_status (status,
				_("No articles retrieved for group %s"),
				group_get_readable_name(task_h->group));
			retval = 0;
		}
	}


	/* `body_index' remembers the last body we downloaded, so that if
	   this task has to be rerun, it resumes at the right place instead
	   of downloading the bodies all over again.*/
	if (retval==-1
		&& task_h->download_bodies
		&& task_h->body_index<task_h->articles->len)
	{
		int val;

		/* refresh the ui */
		status_item_emit_init_steps (status, task_h->articles->len);
		status_item_emit_set_step (status, task_h->body_index);

		/* try downloading the bodies */
		val = nntp_download_bodies (status,
		                            task_h->group,
		                            task->sock,
		                            &task->hint_abort,
		                            task_h->articles,
		                            &task_h->body_index);
		if (val)
			retval = val;
	}

	if (retval==-1)
	{
		/* add the new articles to the Group */
		group_add_articles (task_h->group, task_h->articles);
		g_ptr_array_set_size (task_h->articles, 0);

		/* update the UI, if necessary */
		if (active)
			articlelist_reload ();

		/* cleanup */
		status_item_emit_progress (status, 0);
	}

	/* if we made it this far without fail, return success. */
	if (retval==-1)
		retval = 0;

	/* cleanup */
	if (reffed)
		group_unref_articles (task_h->group, STATUS_ITEM(status));

	return retval;
}

/*****
******
*****/

static int
task_get_header_range (TaskHeaders   * task,
                       gulong        * setme_low,
		       gulong        * setme_high,
		       gulong        * setme_first,
		       gulong        * setme_last,
		       gulong        * setme_qty,
		       const char   ** setme_progress_fmt)
{
	const gboolean * const abort = &TASK(task)->hint_abort;
	PanSocket *sock = TASK(task)->sock;
	Server *server = TASK(task)->server;
	Group *group = TASK_HEADERS(task)->group;
	HeaderDownloadType dl_type = TASK_HEADERS(task)->download_type;
	const char * progress_fmt;
	gulong total;
	gulong first;
	gulong last;
	gulong low;
	gulong high;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);
	g_return_val_if_fail (server!=NULL, -1);
	g_return_val_if_fail (group!=NULL, -1);

	/* change to the right group */
	total = 0;
	first = 0;
	last = 0;
	if (nntp_get_group_info (STATUS_ITEM(task),
	                         sock,
				 group->name,
				 &total,
				 &first,
				 &last,
				 abort))
		return -1;

	/* figure out which articles to ask for, based on the first/last
	 * returned by the server and the dl_type of this task.
	 */
	if (dl_type==HEADERS_NEW)
	{
		gulong p_last = group->article_high;
		if (p_last!=0 && p_last==last)
		{
			status_item_emit_status (STATUS_ITEM(task),
				_("No new articles in group \"%s\""),
			        group_get_readable_name(group));
		}

		low = p_last!=0 ? p_last : first;
		high = last;
		progress_fmt = _("Getting new Article Header %lu of %lu");
	}
	else if (dl_type==HEADERS_SAMPLE)
	{
		const int qty = MIN (total, task->sample_size);
		status_item_emit_status (
			STATUS_ITEM(task), _("Sampling %d articles"), qty);
		low = last - qty;
		high = last;
		total = high - low;
		progress_fmt = _("Sampling Article Header %lu of %lu");
	}
	else if (dl_type==HEADERS_ALL)
	{
		low = first;
		high = last;
		progress_fmt = _("Getting Article Header %lu of %lu");
	}
	else
	{
		pan_warn_if_reached ();
		low = 0;
		high = 0;
		progress_fmt = _("You've found the secret message!!");
	}

	*setme_low = low;
	*setme_high = high;
	*setme_first = first;
	*setme_last = last;
	*setme_qty = total;
	*setme_progress_fmt = progress_fmt;
	return 0;
}


static gchar*
task_headers_describe (const StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

	switch (TASK_HEADERS(item)->download_type)
	{
		case HEADERS_ALL: action=_("all headers"); break;
		case HEADERS_NEW: action=_("new headers"); break;
		case HEADERS_SAMPLE: action=_("sample"); break;
		default: pan_warn_if_reached(); action=_("BUG IN CODE"); break;
	}

	return g_strdup_printf ( _("Getting %s for \"%s\""),
		action,
		group_get_readable_name(TASK_HEADERS(item)->group));
}
